@@ -34,6 +34,7 @@ import           Client.Commands.WordCompletion
 import           Client.Configuration
 import           Client.Mask
 import           Client.Message
+import           Client.Image.PackedImage (imageText)
 import           Client.State
 import           Client.State.Channel
 import           Client.State.DCC
@@ -42,14 +43,15 @@ import           Client.State.Extensions
 import           Client.State.Focus
 import           Client.State.Network
 import           Client.State.Window
+import           Client.UserHost
 import           Control.Applicative
 import           Control.Concurrent.Async (cancel)
-import           Control.Exception (displayException, try)
+import           Control.Exception (displayException, try, SomeException)
 import           Control.Lens
 import           Control.Monad
 import           Data.Foldable
 import           Data.HashSet (HashSet)
-import           Data.List (nub, (\\), elem)
+import           Data.List (nub, (\\))
 import           Data.List.NonEmpty (NonEmpty((:|)))
 import           Data.List.Split
 import qualified Data.HashMap.Strict as HashMap
@@ -57,6 +59,8 @@ import qualified Data.Map as Map
 import           Data.Maybe (fromMaybe)
 import           Data.Text (Text)
 import qualified Data.Text as Text
+import qualified Data.Text.Lazy as LText
+import qualified Data.Text.Lazy.IO as LText
 import           Data.Time
 import           Irc.Commands
 import           Irc.Identifier
@@ -218,7 +222,7 @@ tabCompletion ::
   ClientState      {- ^ client state   -} ->
   IO CommandResult {- ^ command result -}
 tabCompletion isReversed st =
-  case snd $ clientLine st of
+  case dropWhile (' ' ==) $ snd $ clientLine st of
     '/':command -> executeCommand (Just isReversed) command st
     _           -> nickTabCompletion isReversed st
 
@@ -422,6 +426,12 @@ commandsList =
       "Show the TLS certificate for the current connection.\n"
     $ NetworkCommand cmdCert noNetworkTab
 
+  , Command
+      (pure "dump")
+      (simpleToken "filename")
+      "Dump current buffer to file.\n"
+    $ ClientCommand cmdDump simpleClientTab
+
   , Command
       (pure "help")
       (optionalArg (simpleToken "[command]"))
@@ -1183,16 +1193,22 @@ commandsList =
 
   , Command
       (pure "testmask")
-      (liftA2 (,) (simpleToken "[nick!]user@host") (optionalArg (simpleToken "[gecos]")))
+      (liftA2 (,) (simpleToken "[nick!]user@host") (remainingArg "[gecos]"))
       "Test how many local and global clients match a mask.\n"
     $ NetworkCommand cmdTestmask simpleNetworkTab
 
   , Command
       (pure "masktrace")
-      (liftA2 (,) (simpleToken "[nick!]user@host") (optionalArg (simpleToken "[gecos]")))
+      (liftA2 (,) (simpleToken "[nick!]user@host") (remainingArg "[gecos]"))
       "Outputs a list of local users matching the given masks.\n"
     $ NetworkCommand cmdMasktrace simpleNetworkTab
 
+  , Command
+      (pure "trace")
+      (optionalArg (liftA2 (,) (simpleToken "[server | nick]") (optionalArg (simpleToken "[server]"))))
+      "Outputs a list users on a server.\n"
+    $ NetworkCommand cmdTrace simpleNetworkTab
+
   , Command
       (pure "map")
       (pure ())
@@ -1551,6 +1567,25 @@ cmdHelp st mb = commandSuccess (changeSubfocus focus st)
   where
     focus = FocusHelp (fmap Text.pack mb)
 
+-- | Implementation of @/dump@. Writes detailed contents of focused buffer
+-- to the given filename.
+cmdDump :: ClientCommand String
+cmdDump st fp =
+  do res <- try (LText.writeFile fp (LText.unlines outputLines))
+     case res of
+       Left e  -> commandFailureMsg (Text.pack (displayException (e :: SomeException))) st
+       Right{} -> commandSuccess st
+
+  where
+    focus = view clientFocus st
+    msgs  = preview (clientWindows . ix focus . winMessages) st
+    outputLines =
+      case msgs of
+        Nothing  -> []
+        Just wls -> convert [] wls
+    convert acc Nil = acc
+    convert acc (wl :- wls) = convert (views wlFullImage imageText wl : acc) wls
+
 -- | Tab completion for @/splits[+]@. When given no arguments this
 -- populates the current list of splits, otherwise it tab completes
 -- all of the currently available windows.
@@ -1811,14 +1846,24 @@ cmdTestline cs st mask =
   do sendMsg cs (ircTestline (Text.pack mask))
      commandSuccess st
 
-cmdTestmask :: NetworkCommand (String, Maybe String)
+cmdTestmask :: NetworkCommand (String, String)
 cmdTestmask cs st (mask, gecos) =
-  do sendMsg cs (ircTestmask (Text.pack mask) (maybe "" Text.pack gecos))
+  do sendMsg cs (ircTestmask (Text.pack mask) (Text.pack gecos))
      commandSuccess st
 
-cmdMasktrace :: NetworkCommand (String, Maybe String)
+cmdMasktrace :: NetworkCommand (String, String)
 cmdMasktrace cs st (mask, gecos) =
-  do sendMsg cs (ircMasktrace (Text.pack mask) (maybe "*" Text.pack gecos))
+  do sendMsg cs (ircMasktrace (Text.pack mask) (Text.pack gecos))
+     commandSuccess st
+
+cmdTrace :: NetworkCommand (Maybe (String, Maybe String))
+cmdTrace cs st args =
+  do let argsList =
+           case args of
+            Nothing           -> []
+            Just (x, Nothing) -> [x]
+            Just (x, Just y)  -> [x, y]
+     sendMsg cs (rawIrcMsg "TRACE" (map Text.pack argsList))
      commandSuccess st
 
 cmdAway :: NetworkCommand String
@@ -2081,8 +2126,8 @@ cmdReconnect :: ClientCommand ()
 cmdReconnect st _
   | Just network <- views clientFocus focusNetwork st =
 
-      do tm <- getCurrentTime
-         st' <- addConnection 0 (Just tm) Nothing network =<< abortNetwork network st
+      do let tm = preview (clientConnection network . csLastReceived . folded) st
+         st' <- addConnection 0 tm Nothing network =<< abortNetwork network st
          commandSuccess
            $ changeFocus (NetworkFocus network) st'
 
@@ -2309,9 +2354,9 @@ commandNameCompletion isReversed st =
   do guard (cursorPos == n)
      clientTextBox (wordComplete plainWordCompleteMode isReversed [] possibilities) st
   where
-    n = length leadingPart
+    n = length white + length leadingPart
     (cursorPos, line) = clientLine st
-    leadingPart = takeWhile (/=' ') line
+    (white, leadingPart) = takeWhile (' ' /=) <$> span (' '==) line
     possibilities = Text.cons '/' <$> commandNames
     commandNames = keys commands
                 ++ keys (view (clientConfig . configMacros) st)
