@@ -325,14 +325,21 @@ recordChannelMessage network channel msg st
       , rendNicks       = HashSet.fromList (channelUserList network channel' st)
       , rendMyNicks     = highlights
       , rendPalette     = clientPalette st
+      , rendAccounts    = accounts
       }
 
     -- on failure returns mempty/""
-    possibleStatusModes     = view (clientConnection network . csStatusMsg) st
+    cs = st ^?! clientConnection network
+    possibleStatusModes     = view csStatusMsg cs
     (statusModes, channel') = splitStatusMsgModes possibleStatusModes channel
     importance              = msgImportance msg st
     highlights              = clientHighlightsNetwork network st
 
+    accounts =
+      if view (csSettings . ssShowAccounts) cs
+      then view csUsers cs
+      else HashMap.empty
+
 
 recordLogLine ::
   ClientMessage {- ^ message      -} ->
@@ -678,7 +685,7 @@ Right urlPattern =
   compile
     defaultCompOpt
     defaultExecOpt{captureGroups=False}
-    "https?://([[:alnum:]-]+\\.)*([[:alnum:]-]+)(:[[:digit:]]+)?(/[-0-9a-zA-Z$_.+!*'(),%?=:@/;~#]*)?|\
+    "https?://([[:alnum:]-]+\\.)*([[:alnum:]-]+)(:[[:digit:]]+)?(/[-0-9a-zA-Z$_.+!*'(),%?&=:@/;~#]*)?|\
     \<https?://[^>]*>|\
     \\\(https?://[^\\)]*\\)"
 
