@@ -1,4 +1,4 @@
-{-# Language TemplateHaskell, OverloadedStrings, BangPatterns #-}
+{-# Language BlockArguments, TemplateHaskell, OverloadedStrings, BangPatterns #-}
 
 {-|
 Module      : Client.State.Network
@@ -47,12 +47,6 @@ module Client.State.Network
   , csMessageHooks
   , csAuthenticationState
 
-  -- * User information
-  , UserAndHost(..)
-  , uhUser
-  , uhHost
-  , uhAccount
-
   -- * Cross-message state
   , Transaction(..)
 
@@ -83,6 +77,9 @@ import qualified Client.Authentication.Ecdsa as Ecdsa
 import           Client.Configuration.ServerSettings
 import           Client.Network.Async
 import           Client.State.Channel
+import           Client.UserHost
+import           Client.Hook (MessageHook)
+import           Client.Hooks (messageHooks)
 import           Control.Lens
 import           Data.HashMap.Strict (HashMap)
 import qualified Data.HashMap.Strict as HashMap
@@ -125,7 +122,7 @@ data NetworkState = NetworkState
   , _csUsers        :: !(HashMap Identifier UserAndHost) -- ^ user and hostname for other nicks
   , _csModeCount    :: !Int -- ^ maximum mode changes per MODE command
   , _csNetwork      :: !Text -- ^ name of network connection
-  , _csMessageHooks :: ![Text] -- ^ names of message hooks to apply to this connection
+  , _csMessageHooks :: ![MessageHook] -- ^ names of message hooks to apply to this connection
   , _csAuthenticationState :: !AuthenticateState
 
   -- Timing information
@@ -135,7 +132,6 @@ data NetworkState = NetworkState
   , _csLastReceived :: !(Maybe UTCTime) -- ^ time of last message received
   , _csCertificate  :: ![Text]
   }
-  deriving Show
 
 -- | State of the authentication transaction
 data AuthenticateState
@@ -146,14 +142,6 @@ data AuthenticateState
   | AS_ExternalStarted    -- ^ EXTERNAL mode initiated
   deriving Show
 
--- | Pair of username and hostname. Empty strings represent missing information.
-data UserAndHost = UserAndHost
-  { _uhUser    :: {-# UNPACK #-}!Text -- ^ username
-  , _uhHost    :: {-# UNPACK #-}!Text -- ^ hostname
-  , _uhAccount :: {-# UNPACK #-}!Text -- ^ services account
-  }
-  deriving Show
-
 -- | Status of the ping timer
 data PingStatus
   = PingSent !UTCTime -- ^ ping sent at given time, waiting for pong
@@ -178,7 +166,6 @@ data Transaction
   deriving Show
 
 makeLenses ''NetworkState
-makeLenses ''UserAndHost
 makePrisms ''Transaction
 makePrisms ''PingStatus
 makePrisms ''TimedAction
@@ -271,7 +258,7 @@ newNetworkState network settings sock ping = NetworkState
   , _csModeCount    = 3
   , _csUsers        = HashMap.empty
   , _csNetwork      = network
-  , _csMessageHooks = view ssMessageHooks settings
+  , _csMessageHooks = buildMessageHooks (view ssMessageHooks settings)
   , _csAuthenticationState = AS_None
   , _csPingStatus   = ping
   , _csLatency      = Nothing
@@ -280,6 +267,10 @@ newNetworkState network settings sock ping = NetworkState
   , _csCertificate  = []
   }
 
+buildMessageHooks :: [HookConfig] -> [MessageHook]
+buildMessageHooks = mapMaybe \(HookConfig name args) ->
+  do hookFun <- HashMap.lookup name messageHooks
+     hookFun args
 
 -- | Used for updates to a 'NetworkState' that require no reply.
 --
@@ -304,10 +295,17 @@ applyMessage' msgWhen msg cs =
     Ping args -> ([ircPong args], cs)
     Pong _    -> noReply $ doPong msgWhen cs
     Join user chan acct ->
-           noReply
-         $ recordUser user acct
+         ( reply
+         , recordUser user acct
          $ overChannel chan (joinChannel (userNick user))
-         $ createOnJoin user chan cs
+         $ createOnJoin user chan cs )
+     where
+       showAccounts = view (csSettings . ssShowAccounts) cs
+       reply
+         | userNick user == view csNick cs =
+              ircMode chan [] :
+              [ircWho [idText chan, "%tuhna,616"] | showAccounts ]
+         | otherwise = []
 
     Account user acct ->
            noReply
@@ -691,7 +689,9 @@ selectCaps ::
   NetworkState         {- ^ network state  -} ->
   [(Text, Maybe Text)] {- ^ server caps    -} ->
   [Text]               {- ^ caps to enable -}
-selectCaps cs offered = supported `intersect` Map.keys capMap
+selectCaps cs offered = (supported `intersect` Map.keys capMap)
+                        `union`
+                        view (csSettings . ssCapabilities) cs
   where
     capMap = Map.fromList offered
 
